/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file    comms_queue.h
 *
 *  @brief   Double linked communications queue.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/

#ifndef _COMMS_QUEUE_H_
#define _COMMS_QUEUE_H_

#include "records.h"

typedef struct _QueueItem_t
{
  CommRecord_t *comm; /* Communication info. */
  struct _QueueItem_t  *next; /* Linked by time. */
  struct _QueueItem_t  *prev; /* Linked by time. When it's the first item, also used */
                                /* to point to the last item of the time queue. */
  struct _QueueItem_t  *thread_next; /* Linked by thread. */
  struct _QueueItem_t  *thread_prev; /* Linked by thread. When it's the first item,  */
                                       /* also used to point to the last item of the   */
                                       /* thread queue.                                */
} QueueItem_t;

typedef struct
{
  QueueItem_t *thread_queue[ MAX_THREAD ]; /* Thread queues */
  QueueItem_t *first_item;                 /* Global queue  */
                                           /* Is also the first time. */
  uint32_t    num_comms;                   /* Total number of communications */
} ThreadCommsQueue_t;


/* QueueItem_t operations */
QueueItem_t *CreateItem( CommRecord_t *comm );
void FreeItem( QueueItem_t *queue);

ThreadCommsQueue_t *CreateCommsQueue( void );
void FreeCommsQueue( ThreadCommsQueue_t *tq );

/**
 * Returns the first communication of a thread. Doesn't dequeue. Increments
 * global communications number.
 * Thread Queue function.
 *
 * @param queue  (In) Double linked communications queue.
 * @param thread (In) Thread of the sender
 * @return Pointer to the communication or NULL if void queue.
 */
CommRecord_t *FirstComm( ThreadCommsQueue_t *queue, uint32_t thread );

/**
 * Enqueues a communication by its sender thread. Creates QueueItem_t record.
 * Thread Queue function.
 *
 * @param queue  (In) Double linked communications queue.
 * @param thread (In) Thread of the sender.
 * @param comm   (In) Record to enqueue.
*/
void QueueComm( ThreadCommsQueue_t *queue,
                uint32_t            thread,
                CommRecord_t       *comm );

/**
 * Returns the first communication item of a thread.
 * Thread Queue function.
 *
 * @param queue  (In) Double linked communications queue.
 * @param thread (In) Thread of the sender.
 * @return Pointer to the first item or NULL if void queue.
 */
QueueItem_t *FirstItem( ThreadCommsQueue_t *queue, uint32_t thread );

/**
 * Returns the prev communication item of a thread.
 * Thread Queue function.
 *
 * @param queue (In) Double linked communications queue.
 * @param item  (In) Current item.
 * @return Pointer to the prev item or NULL if none.
 */
QueueItem_t *PrevItem( ThreadCommsQueue_t *queue, QueueItem_t *item );

/**
 * Returns the next communication item of a thread.
 * Thread Queue function.
 *
 * @param queue (In) Double linked communications queue.
 * @param item  (In) Current item.
 * @return Pointer to the next item or NULL if none.
 */
QueueItem_t *NextItem( ThreadCommsQueue_t *queue, QueueItem_t *item );

CommRecord_t *SearchComm( ThreadCommsQueue_t *queue,
                            uint32_t            thread,
                            CommRecord_t       *comm );
/**
 * Dequeues the first communication record. Deletes QueueItem_t record.
 * Decrements global comms number.
 * Thread Queue function.
 *
 * @param queue  (In) Double linked communications queue.
 * @param thread  (In) Thread of the sender.
 */
void DequeueComm( ThreadCommsQueue_t *queue, uint32_t thread );

/* Queue operations by time. */

QueueItem_t *FirstGlobalItem( ThreadCommsQueue_t *queue );
QueueItem_t *NextGlobalItem( QueueItem_t *item );

/* Get info */
CommRecord_t *GetCommInfo( QueueItem_t *item );
CommRecord_t *FirstGlobalComm( ThreadCommsQueue_t *queue );
unsigned int Log_Unmatched_Comms( ThreadCommsQueue_t *com_queue, const char *msg );

int TestQueue( ThreadCommsQueue_t *queue );

#endif /* _COMMS_QUEUE_H_ */
