/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file    comms_queue.c
 *
 *  @brief  Implements Double linked communications queue.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "comms_queue.h"
#include "log.h"
#include "statistics.h"

extern FILE *err_fd;
extern Statistics_t stats;

/* QueueItem_t functions */
QueueItem_t *CreateItem( CommRecord_t *com )
{
  QueueItem_t *new;

  new = ( QueueItem_t * )malloc( sizeof( QueueItem_t ));
  if ( new == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  new->comm        = com;
  new->next        = NULL;
  new->prev        = NULL;
  new->thread_prev = NULL;
  new->thread_next = NULL;

  return new;
}


void FreeItem( QueueItem_t *queue)
{
  free( queue->comm );
  free( queue );
  queue = NULL;
}


ThreadCommsQueue_t *CreateCommsQueue( void )
{
  ThreadCommsQueue_t *tq;
  uint32_t ii;

  tq = ( ThreadCommsQueue_t * )malloc( sizeof( ThreadCommsQueue_t ));
  if ( tq == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  tq->first_item = NULL;
  tq->num_comms = 0;
  for ( ii = 0; ii < MAX_THREAD; ii++ )
    tq->thread_queue[ ii ] = NULL;

  return tq;
}


void FreeCommsQueue( ThreadCommsQueue_t *tq )
{
  if ( tq->num_comms == 0 )
  {
    free( tq );
    tq = NULL;
  }
}


QueueItem_t *FirstGlobalItem( ThreadCommsQueue_t *queue )
{
  return queue->first_item;
}


QueueItem_t *NextGlobalItem( QueueItem_t *item )
{
  if ( item != NULL )
    return item->next;
  else
    return NULL;
}


CommRecord_t *GetCommInfo( QueueItem_t *item )
{
  if ( item != NULL )
    return item->comm;
  else
    return NULL;
}


CommRecord_t *FirstGlobalComm( ThreadCommsQueue_t *queue )
{
  return GetCommInfo( FirstGlobalItem( queue ));
}


CommRecord_t *FirstComm( ThreadCommsQueue_t *queue, uint32_t thread )
{
  CommRecord_t *cr;

  if ( queue->thread_queue[ thread ] == NULL )
    cr = NULL;
  else
    cr = (queue->thread_queue[ thread ])->comm;

  return ( cr );
}


void QueueComm( ThreadCommsQueue_t *queue,
                uint32_t            thread,
                CommRecord_t       *comm )
{
  QueueItem_t *new_item, *last_item;

  new_item = CreateItem( comm );

  /* Insert in thread queue. */
  if ( queue->thread_queue[ thread ] == NULL ) /* void queue */
  {
    queue->thread_queue[ thread ] = new_item;
    new_item->thread_prev = new_item;
  }
  else /* first->thread_prev used as last. */
  {
    last_item = queue->thread_queue[ thread ]->thread_prev;
    last_item->thread_next = new_item;
    new_item->thread_prev = last_item;
    queue->thread_queue[ thread ]->thread_prev = new_item;
  }

  /* Insert in global queue. */
  if ( queue->first_item == NULL ) /* void queue */
  {
    queue->first_item = new_item;
    new_item->prev = new_item;
  }
  else /* first_item->prev used as last. */
  {
    last_item = queue->first_item->prev;
    last_item->next = new_item;
    new_item->prev  = last_item;
    queue->first_item->prev = new_item;
  }
  new_item->next = NULL;
  queue->num_comms++;
}


QueueItem_t *FirstItem( ThreadCommsQueue_t *queue, uint32_t thread )
{
  return ( queue->thread_queue[ thread ] );
}


QueueItem_t *NextItem( ThreadCommsQueue_t *queue, QueueItem_t *item )
{
  return ( item->thread_next );
}


QueueItem_t *PrevItem( ThreadCommsQueue_t *queue, QueueItem_t *item )
{
  return ( item->thread_prev );
}


int TestQueue( ThreadCommsQueue_t *queue )
{
  int count = 0;
  QueueItem_t *item;

  item = FirstGlobalItem( queue );
  while ( item != NULL )
  {
    count++;
    item = NextGlobalItem( item );
  }

  return count;
}


CommRecord_t *SearchComm( ThreadCommsQueue_t *queue,
                          uint32_t            thread,
                          CommRecord_t       *comm )
{
  CommRecord_t *returned_comm = NULL;
  CommRecord_t *current_comm;
  QueueItem_t *current_item;

  returned_comm = NULL;
  
  current_item = FirstItem( queue, thread );
  while (( current_item != NULL ) && ( returned_comm == NULL ))
  {
    current_comm = current_item->comm;
    if (( !current_comm->matched ) && AreCommsEqual( current_comm, comm ))
      returned_comm = current_comm;

    current_item = NextItem( queue, current_item );        
  }

  return returned_comm;
}


void DequeueComm( ThreadCommsQueue_t *queue, uint32_t thread )
{
  QueueItem_t *first;

  if ( queue->thread_queue[ thread ] != NULL )
  {
    first = queue->thread_queue[ thread ];

    if ( first->thread_next != NULL )
    {
      first->thread_next->thread_prev = first->thread_prev; /* LAST */
      queue->thread_queue[ thread ] = first->thread_next;
    }
    else
      queue->thread_queue[ thread ] = NULL;

    /* Relink global */
    if ( queue->first_item == first ) /* FIRST RECORD */
    {
      if ( first->next != NULL )
      {
        first->next->prev = first->prev; /*LAST*/
        queue->first_item = first->next;
      }
      else
      {
        queue->first_item = NULL;
      }
    }
    else if (queue->first_item->prev == first) /* IS LAST */
    {
      queue->first_item->prev = first->prev;
      first->prev->next = NULL;
    }
    else /* in the middle */
    {
      first->prev->next = first->next;
      first->next->prev = first->prev;
    }

    FreeItem( first );

    /* decrement comms number */
    queue->num_comms--;
  }
}


unsigned int Log_Unmatched_Comms( ThreadCommsQueue_t *com_queue, const char* msg )
{
  QueueItem_t *item;
  CommRecord_t *cm;
  char line[ MAX_LINE ];
  unsigned int unmatched = 0;

  item = FirstGlobalItem( com_queue );
  while (item != NULL)
  {
    cm = item->comm;
    
    if ( !cm->matched )
    {
      Write_Log( err_fd, ALWAYS, (const char *)msg );
      sprintf( line, FORMAT_COMMS_RECORD,
                   cm->snd_cpu, cm->snd_ptask, cm->snd_task, cm->snd_thread,
                   (unsigned long long int)cm->snd_time, (unsigned long long int)cm->snd_time,
                   cm->rcv_cpu, cm->rcv_ptask, cm->rcv_task, cm->rcv_thread,
                   (unsigned long long int)cm->rcv_time, (unsigned long long int)cm->rcv_time,
                   cm->size, cm->tag );
      Write_Log( err_fd, ALWAYS, line );
      unmatched++;
    }
    
    item = NextGlobalItem( item );
  }

  return unmatched;
}
