/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file  events_queue.c
 *
 *  @brief  Events queue.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/

#include <stdio.h>
#include <stdlib.h>

#include "types.h"
#include "events_queue.h"
#include "statistics.h"

extern FILE *err_fd;
extern Statistics_t stats;

/* EvQueueItem_t functions */
EvQueueItem_t *CreateCopyEvItem( EventRecord_t *ev )
{
  EvQueueItem_t *eqi;
  EventRecord_t *ex;

  /* Create an item */
  eqi = ( EvQueueItem_t * )malloc( sizeof( EvQueueItem_t ));
  if ( eqi == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }
  
  /* Create void event */
  ex  = CreateEventRecord( 0,0,0,0,0,0,0 );
  
  /* Fill it with ev info */
  CopyEventRecord( ev, ex );
  
  eqi->ev   = ex;
  eqi->next = NULL;

  return eqi;
}


EvQueueItem_t *CreateEvItem( EventRecord_t *ev )
{
  EvQueueItem_t *eqi;

  /* Create an item */
  eqi = ( EvQueueItem_t * )malloc( sizeof( EvQueueItem_t ));
  if ( eqi == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  /* Directly copy ev addres */
  eqi->ev   = ev;
  eqi->next = NULL;

  return eqi;
}


void FreeEvItem( EvQueueItem_t *item )
{
  if ( item != NULL )
  {
    free( item->ev );
    item->ev = NULL;
  }

  free( item );
  item = NULL;
}


void GetEvent( EvQueueItem_t *item, EventRecord_t **ev )
{
  if (( item != NULL ) && ( ev != NULL ))
    *ev = item->ev;
}


/* EventsQueue_t functions */
EventsQueue_t *CreateEventsQueue( void )
{
  EventsQueue_t *eq;

  eq = ( EventsQueue_t * )malloc( sizeof( EventsQueue_t ));
  if (eq == NULL)
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  eq->first = NULL;
  eq->last  = NULL;
  eq->num_events = 0;

  return eq;
}


void FreeEventsQueue( EventsQueue_t *eq )
{
  EvQueueItem_t *aux;

  if ( eq != NULL )
  {
    while ( eq->num_events > 0 )
    {
      aux = eq->first;
      eq->first = (eq->first)->next;
      FreeEvItem( aux );
      eq->num_events--;
    }
  }
  
  free( eq );
  eq = NULL;
}

/* Queue er */
void InsertEvent( EventsQueue_t *eq, EventRecord_t *er )
{
  EvQueueItem_t *eqi;

  if ( er->cpu > 500)
    abort();

  eqi = CreateEvItem( er );
  if ( eq->num_events == 0 )
  {
    eq->first = eqi;
    eq->last  = eqi;
  }
  else
  {
    eq->last->next = eqi;
    eq->last = eqi;
  }
  eq->num_events++;

  stats.queue[ QEVENTS ].queued++;
  stats.queue[ QEVENTS ].total_sum_queue += eq->num_events;
}

/* Stack er */
void PushEvent( EventsQueue_t *eq, EventRecord_t *er, bool_t create )
{
  EvQueueItem_t *eqi;

  if ( create )
    eqi = CreateCopyEvItem( er );
  else
    eqi = CreateEvItem( er );
  
  if ( eq->num_events == 0 )
  {
    eq->first = eqi;
    eq->last  = eqi;
  }
  else
  {
    eqi->next = eq->first;
    eq->first = eqi;
  }
  eq->num_events++;
}


bool_t IsEmptyEventsQueue( EventsQueue_t *eq )
{
  return ( eq->num_events == 0 );
}


/* returns event address */
void GetFirstEvent( EventsQueue_t *eq, EventRecord_t **ev )
{
  if ( !IsEmptyEventsQueue( eq ))
    GetEvent( eq->first, ev );
}

/* extracts but doesnt free event */
void DequeueFirstEvent( EventsQueue_t *eq ) 
{
  EvQueueItem_t *aux;

  if ( !IsEmptyEventsQueue( eq ))
  {
    aux = eq->first;
    eq->first->ev = NULL; /* CAUTION: MAY LOOSE MEMORY */
    eq->first = eq->first->next;
    eq->num_events--;
    FreeEvItem( aux );
  }
}

/* gets first and extracts */
EventRecord_t *FirstEventsQueue( EventsQueue_t *eq )
{
  EventRecord_t *ret;
  EvQueueItem_t *aux;
  
  ret = NULL;
  
  if ( !IsEmptyEventsQueue( eq ))
  {
    ret = eq->first->ev;
    aux = eq->first;
    eq->first->ev = NULL;
    eq->first = eq->first->next;
    eq->num_events--;
    FreeEvItem( aux );
  }

  return ret;
}


EntryStack_t *CreateEntryStack( void )
{
  EntryStack_t *stack;
  int ii;

  stack = ( EntryStack_t * )malloc( sizeof( EntryStack_t ));
  if (stack == NULL)
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  for ( ii = 0; ii < MAX_THREAD; ii++ )
    stack->entry[ii] = NULL;

  return stack;
}


void PushEventStack( EntryStack_t *stack, 
                      EventRecord_t *er,
                      uint32_t level,
                      bool_t create )
{
  if ( stack->entry[ level ] == NULL )
    stack->entry[ level ] = CreateEventsQueue( );

  PushEvent( stack->entry[ level ], er, create );
}


bool_t IsEmptyEventsStackQueue( EntryStack_t *stack, uint32_t level )
{
  bool_t ret = FALSE;

  if ( stack == NULL )
    ret = TRUE;
  else if ( stack->entry[ level ] == NULL )
    ret = TRUE;

  return ret;
}


void FreeEntryStack( EntryStack_t *stack )
{
  int ii;

  for ( ii = 0; ii < MAX_THREAD; ii++ )
  {
    if ( stack->entry[ii] != NULL )
    {
      if ( stack->entry[ ii ]->first != NULL )
      {
        fprintf(err_fd,"ENTRY_STACK:ERROR: queued event on free!!!");
      }
      
      FreeEventsQueue( stack->entry[ii] );
    }
  }
  free(stack);
  stack = NULL;
}
