/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file  otf2prv.c
 *
 *  @brief  Translates traces from OTF format to Paraver format.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/

/* ---[ SYSTEM INCLUDES ]---------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>

/* ---[ OTF2PRV INCLUDES ]--------------------------------------------------- */
#include "reader.h"
#include "parse_parameters.h"
#include "tokens.h"
#include "events_queue.h"
#include "comms_queue.h"
#include "otf_symbolic_info.h"
#include "log.h"
#include "paraver_header.h"
#include "statistics.h"

/* ---[ CONSTANTS ]---------------------------------------------------------- */
#define ERROR -1
#define OK     0
#define ERROR_SEND    "ERROR: Unmatched Send "
#define ERROR_RECEIVE "ERROR: Unmatched Recv "

/* ---[ GLOBAL VARIABLES ]--------------------------------------------------- */
TranslatorOptions_t *parameters;
FILE *prv_fd, *pcf_fd, *err_fd, *tmp_fd, *comm_fd;
char *tmp_name, *comm_name;
unsigned long long int trace_max_time;

ParaverHeader_t *header;

ThreadCommsQueue_t *comms;
ThreadCommsQueue_t *unmatched_receives;
EntryStack_t       *function_stack;

void Dump_Communications (ThreadCommsQueue_t *);

#ifdef VAMPIR
EntryStack_t       *enter_events_stack;
uint64_t             ticks_per_second;
uint64_t             unit_factor;
#endif

Statistics_t stats;

/* ---[ LOCAL HEADERS ]------------------------------------------------------ */
static bool_t Initialize_Translation( TranslatorOptions_t *par );
static bool_t Exist_File( char *otf_name );
static void Merge_Files( FILE *prv_file, FILE *events_file, FILE *comms_file );
static void   Finish( TranslatorOptions_t *par );

static bool_t Open_Write_Files ( TranslatorOptions_t *par );
static void   Create_Queues    ( TranslatorOptions_t *par );
static void   Free_Queues      ( TranslatorOptions_t *par );
static void   Close_Write_Files( TranslatorOptions_t *par );


/* ---[ FIRST LEVEL ]-------------------------------------------------------- */
void Signal_SIGUSR( int sig )
{
  exit( 1 );
}

static bool_t Initialize_Translation( TranslatorOptions_t *par )
{
  bool_t initialized;

  initialized = Open_Write_Files( parameters );
  if ( initialized )
  {
    Write_Log( log_fd, SHADOW, "Initializating translator.\n" );
    Create_Queues( parameters );
    header = Create_Paraver_Header();
    Init_Statistics( &stats );
  }

  return initialized;
}

static bool_t Exist_File( char *otf_name )
{
  bool_t exist;
  int fd;

  exist = FALSE;
  if ( otf_name != NULL )
  {
    fd = open( otf_name, O_RDONLY );
    if ( fd < 0 )
    {
      perror( otf_name );
    }
    else
    {
      close( fd );
      exist = TRUE;
    }
  }

  return exist;
}


static void Merge_Files( FILE *prv_file, FILE *events_file, FILE *comms_file )
{
  char line_event[ MAX_LINE ];
  char line_comm[ MAX_LINE ];
  int aux;
  uint64_t time_event, time_comm;
  size_t size_line;
	bool_t last_is_event = TRUE;
  
  Write_Log( log_fd, SHADOW, "Merging files.\n" );

  size_line = sizeof( char ) * MAX_LINE;

  fclose( events_file );
  fclose( comms_file );  
  events_file = fopen( tmp_name,  "r" );
  comms_file  = fopen( comm_name, "r" );

  /* Initial reads */
  if( fgets( line_event, size_line, events_file ) != NULL )
    sscanf( line_event, FORMAT_EVENT_RECORD_REDUCED, &aux, &aux, &aux, &aux, (long long *)&time_event );

  if( fgets( line_comm, size_line, comms_file ) != NULL )
    sscanf( line_comm, FORMAT_COMMS_RECORD_REDUCED, &aux, &aux, &aux, &aux, (long long *)&time_comm );

  /* Merging loop */
  while ( !feof( events_file ) && !feof( comms_file ))
  {

    if ( time_event <= time_comm )
    {
      printf( "%lld <= %lld\n", (long long)time_event, (long long)time_comm );
      fputs( line_event, prv_file );
      if( fgets( line_event, size_line, events_file ) != NULL )
        sscanf( line_event, FORMAT_EVENT_RECORD_REDUCED, &aux, &aux, &aux, &aux, (long long *)&time_event );
    }    
    else
    {
      printf( "%lld > %lld\n", (long long)time_event, (long long)time_comm );
      fputs( line_comm, prv_file );
      if( fgets( line_comm, size_line, comms_file ) != NULL )
        sscanf( line_comm, FORMAT_COMMS_RECORD_REDUCED, &aux, &aux, &aux, &aux, (long long *)&time_comm );
    }
  }

  /* Read in previous loop */
  if ( !feof( events_file ) )
    fputs( line_event, prv_file );

  if ( !feof( comms_file ) )
	{
    fputs( line_comm, prv_file );
  	last_is_event = FALSE;
	}

  /* Tail of files */
  while ( !feof( events_file ) )
  {
    if ( fgets( line_event, size_line, events_file ) != NULL )
		{
      fputs( line_event, prv_file );
  		last_is_event = TRUE;
		}
  }

  while ( !feof( comms_file ) )
  {
    if ( fgets( line_comm, size_line, comms_file ) != NULL )
		{
      fputs( line_comm, prv_file );
			last_is_event = FALSE;
		}
  }

	if ( last_is_event )
	  fprintf( prv_file, "\n" );	
	
  fclose( events_file );
  fclose( comms_file );
  
  remove( tmp_name );
  remove( comm_name );
}


static void Finish( TranslatorOptions_t *par )
{
  Write_Log( log_fd, SHADOW, "Freeing resources..." );

  Close_Write_Files( parameters );
  Free_Queues( parameters );
  Free_Paraver_Header( header );

  Write_Log( log_fd, SHADOW, "OK\n" );
}

/* ---[ SECOND LEVEL ]------------------------------------------------------- */
static bool_t Open_Write_Files( TranslatorOptions_t *par )
{
  bool_t opened_files;
  char *prv_name, *pcf_name, *err_name;

  log_fd = Open_Log( par );
  Write_Log( log_fd, SHADOW, "Opening files..." );
  if ( Exist_File( par->otf_trace_name ))
  {
    err_name = New_File_Name( parameters, SUFFIX_ERR );
    err_fd = fopen( err_name, "w");
    free( err_name );

    if ( !par->option[OPTION_ZIP] )
    {
      prv_name = New_File_Name( parameters, SUFFIX_PRV );
      pcf_name = New_File_Name( parameters, SUFFIX_PCF );
      tmp_name = New_File_Name( parameters, SUFFIX_EVT );
      comm_name = New_File_Name( parameters, SUFFIX_COM );

      prv_fd = fopen( prv_name, "w" );
      pcf_fd = fopen( pcf_name, "w" );
      tmp_fd = fopen( tmp_name, "w" );
      comm_fd = fopen( comm_name, "w" );

      free( prv_name );
      free( pcf_name );
    }
    opened_files = TRUE;
    Write_Log( log_fd, SHADOW, "OK\n" );
  }
  else
  {
    Write_Log( log_fd, SHADOW, "KO!\nError!: Can't open OTF file.\n" );
    opened_files = FALSE;
  }

  return opened_files;
}

static void Create_Queues( TranslatorOptions_t *par )
{
  Write_Log( log_fd, SHADOW, "Allocating queues..." );

  comms  = CreateCommsQueue();
  unmatched_receives = CreateCommsQueue();
  function_stack = CreateEntryStack();
#ifdef VAMPIR
  enter_events_stack = CreateEntryStack();
#endif
  
  Write_Log( log_fd, SHADOW, "OK\n" );
}

static void Free_Queues( TranslatorOptions_t *par )
{
  Write_Log( log_fd, SHADOW, "Freeing queues..." );

  FreeCommsQueue( comms );
  FreeCommsQueue( unmatched_receives );
  FreeEntryStack(function_stack );
#ifdef VAMPIR
  FreeEntryStack( enter_events_stack );
#endif

  Write_Log( log_fd, SHADOW, "OK\n" );
}

static void Close_Write_Files( TranslatorOptions_t *par )
{
  Write_Log( log_fd, SHADOW, "Closing files..." );

  if ( !par->option[OPTION_ZIP] )
  {
    fclose( prv_fd );
    fclose( pcf_fd );
  }

  fclose( err_fd );

  Write_Log( log_fd, SHADOW, "OK\n" );
}

/* ---[ MAIN BODY ]---------------------------------------------------------- */
int main( int argc, char *argv[] )
{
  int ret_code;
  
  signal( SIGUSR1, Signal_SIGUSR );

  ret_code = 0;
  log_fd = Open_Log( NULL );

  if (argc < 2)
    Show_Usage();
  else
  {
    parameters = Create_Options();
    Parse_Options( argc, argv, parameters );
    if ( parameters != NULL )
    {
      if ( parameters->option[ OPTION_HELP ] ||
           !parameters->option[ OPTION_OTF ] )
      {
        Show_Usage();
      }
      else if ( parameters->option[ OPTION_OTF ] )
      {
        if ( Initialize_Translation( parameters ))
        {
          Write_Log( log_fd, SHADOW, "Starting translation.\n" );
          Translate(  parameters->otf_trace_name,
                      Get_EVENTS_Name( parameters ),
                      tmp_fd, pcf_fd );
          Write_Log( log_fd, SHADOW, "Generating PCF file.\n" );
          Generate_PCF( sym_otf_info, pcf_fd );

          Write_Log( log_fd, SHADOW, "Generating PRV file.\n" );
          Generate_Paraver_Header( header, prv_fd );
          Dump_Communications( comms );
          Merge_Files( prv_fd, tmp_fd, comm_fd );
          
          Write_Log( log_fd, SHADOW, "Logging unmatched sends and receives.\n" );
          stats.otf_unmatched_sends    = Log_Unmatched_Comms( comms, ERROR_SEND );
          stats.otf_unmatched_receives = Log_Unmatched_Comms( unmatched_receives, ERROR_RECEIVE );

          Write_Log( log_fd, SHADOW, "Printing statistics.\n" );
          Print_Statistics( &stats, parameters, argc, argv );
					Print_Otf_Symbolic_Info( sym_otf_info   );
					
          Finish( parameters );
        }
        Write_Log( log_fd, SHADOW, "Exiting otf2prv.\n" );
        if ( parameters->option[OPTION_LOG] )
          fclose( log_fd );
      }
    }
    else
    {
      ret_code = -1;
      Write_Log( err_fd, SHADOW, "Problems parsing parameters!!!\n" );
    }
  }

  return ret_code;
}
