/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file    otf_symbolic_info.c
 *
 *  @brief   Implements table to translate symbolic tokens from otf to prv.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "types.h"
#include "otf_symbolic_info.h"

#define DEFAULT_COLORS_PCF 16

extern FILE *err_fd;

static void Partial_Generate_PCF( SymbolicInfo_t *si, FILE *fd, char group, char type);


OtfToken_t *Create_Otf_Token( char token_type,
                              unsigned int token_group,
                              unsigned int token_id,
                              const char *token_name,
                              unsigned int alias_id,
                              unsigned int masked_id )
{
  OtfToken_t *token;

  token = ( OtfToken_t * )malloc( sizeof( OtfToken_t ));
  if ( token == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit(1);
  }

  token->otf_token_type  = token_type;
  token->otf_token_group = token_group;
  token->otf_token_id    = token_id;
  token->otf_token_name  = strdup( token_name );
  token->prv_color       = 0;
  token->alias_id        = alias_id;
  token->masked_id       = masked_id;

  return token;
}


void Print_Otf_Token( OtfToken_t *token, int ii )
{
  if ( token->alias_id == 0)
    fprintf( stdout, "(%i) [%c] otf_group %i otf_token %i (%s) ->  %i \n",
                   ii,
                   token->otf_token_type,
                   token->otf_token_group,
                   token->otf_token_id,
                   token->otf_token_name,
                   token->prv_color );
  else
    fprintf( stdout, "(%i) [%c] otf_group %i otf_token %i / %i (%s) ->  %i \n",
                   ii,
                   token->otf_token_type,
                   token->otf_token_group,
                   token->otf_token_id,
                   token->alias_id,
                   token->otf_token_name,
                   token->prv_color );
}


void Free_Otf_Token( OtfToken_t *token )
{
  if (token != NULL)
    free( token->otf_token_name );
  free( token );
  token = NULL;
}


SymbolicInfo_t *Create_Otf_Symbolic_Info( void )
{
  SymbolicInfo_t *sym_info;

  sym_info = (SymbolicInfo_t *)malloc( sizeof( SymbolicInfo_t ));
  if (sym_info == NULL)
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }
  sym_info->num_tokens = 0;

  return sym_info;
}


void Insert_Otf_Token( OtfToken_t *token, SymbolicInfo_t *sym_info )
{
  memcpy( &sym_info->token_info[ sym_info->num_tokens ],
          token,
          sizeof( OtfToken_t ));
  sym_info->num_tokens++;
}


int compare_tokens( OtfToken_t *t1, OtfToken_t *t2 )
{
  if ( t1->otf_token_id < t2->otf_token_id )
    return -1;
  else if ( t1->otf_token_id > t2->otf_token_id )
    return 1;
	if ( t1->otf_token_type < t2->otf_token_type )
    return -1;
  else if ( t1->otf_token_type > t2->otf_token_type )
    return 1;
	if ( t1->otf_token_group < t2->otf_token_group )
    return -1;
  else if ( t1->otf_token_group > t2->otf_token_group )
    return 1;
  else
    return 0;
}


void Insert_Otf_Info( char         token_type,
                      unsigned int token_group,
                      unsigned int token_id,
                      const char *token_name,
                      SymbolicInfo_t *sym_info,
                      unsigned int masked_id )
{
  OtfToken_t *token;
  unsigned int alias_id = 0;

#ifdef VAMPIR
  if ( token_type == COUNTER_GROUP_TYPE )
  {
    alias_id = token_id;
    token_id = Inserted_Tokens( sym_info );
  }
#endif

  /* PENDIENTE MASKED_ID */
  token = Create_Otf_Token( token_type, token_group, token_id, token_name, alias_id, masked_id );
  Insert_Otf_Token( token, sym_info );
	qsort( sym_info->token_info,
         ( size_t )sym_info->num_tokens,
         sizeof( OtfToken_t ),
         (int (*)(const void *, const void *))compare_tokens );
  /* Free_Otf_Token( token ); */

}


unsigned int Inserted_Tokens( SymbolicInfo_t *sym_info )
{
  return sym_info->num_tokens;
}

int compare_key( OtfToken_Key_t *t1, OtfToken_t *t2 )
{
  if ( t1->otf_token_id < t2->otf_token_id )
    return -1;
  else if ( t1->otf_token_id > t2->otf_token_id )
    return 1;
	if ( t1->otf_token_type < t2->otf_token_type )
    return -1;
  else if ( t1->otf_token_type > t2->otf_token_type )
    return 1;
  else
    return 0;
}

bool_t Get_Otf_Token_Group( SymbolicInfo_t *sym_info, unsigned int token_id,
                                  char type, unsigned int *token_group )
{
  int ii;
  bool_t found = FALSE;
  OtfToken_Key_t key;
  OtfToken_t *tok;
  
  key.otf_token_id   = token_id;
  key.otf_token_type = type;
  
  tok = bsearch( &key,
                 sym_info->token_info,
                 sym_info->num_tokens,
                 sizeof(OtfToken_t),
                 (int (*)(const void *, const void *))compare_key );
  
  if ( tok == NULL )
  {
    fprintf( err_fd, "SYMBOLIC: Token %i not found in symbols table.\n",
                     token_id );
		*token_group = 0;
	}
  else
  {
    found = TRUE;
    *token_group = tok->otf_token_group;    
  }
  return found;
}

static  bool_t Is_MPI( SymbolicInfo_t *sym_info,
                          unsigned int token_value,
                          const char *func_name)
{
  bool_t found = FALSE;
  OtfToken_Key_t key;
  OtfToken_t *tok;
  
  key.otf_token_id   = token_value;
  key.otf_token_type = FUNCTION_TYPE;
  
  tok = bsearch( &key,
                 sym_info->token_info,
                 sym_info->num_tokens,
                 sizeof(OtfToken_t),
                 (int (*)(const void *, const void *))compare_key );
  
  if ( tok != NULL )
  {
    if (!strcmp( tok->otf_token_name, func_name ))
        found = TRUE;
  }
  
  return found;
}


unsigned int Get_Otf_Token_AliasId( SymbolicInfo_t *sym_info,
                                  unsigned int token_id,
                                  char type,
                                  unsigned int *masked_id)
{
  int ii;
  unsigned int alias_id = 0;

  for ( ii = 0; ii < sym_info->num_tokens; ii++ )
  {
    if (( sym_info->token_info[ ii ].alias_id == token_id ) &&
        ( sym_info->token_info[ ii ].otf_token_type == type ))
    {
      alias_id  = sym_info->token_info[ ii ].otf_token_id;
      *masked_id = sym_info->token_info[ ii ].masked_id;
      break;
    }
  }

  return alias_id;
}


void Set_Token_MaskedId( SymbolicInfo_t *sym_info,
                          unsigned int alias_id,
                          unsigned int masked_id)
{
  int ii;

  for ( ii = 0; ii < sym_info->num_tokens; ii++ )
  {
    if ( sym_info->token_info[ ii ].otf_token_id == alias_id )
    {
      sym_info->token_info[ ii ].masked_id = masked_id;
      break;
    }
  }
}


unsigned int Get_Token_MaskedId( SymbolicInfo_t *sym_info,
                                   unsigned int alias_id)
{
  int ii;

  for ( ii = 0; ii < sym_info->num_tokens; ii++ )
  {
    if ( sym_info->token_info[ ii ].otf_token_id == alias_id )
    {
      break;
    }
  }
  
  if ( ii == sym_info->num_tokens )
    return 0;
  else
    return sym_info->token_info[ ii ].masked_id;
}


bool_t Is_MPI_Wait( SymbolicInfo_t *sym_info,
                     unsigned int token_value )
{
/*
  #ifdef VAMPIR
  return (Is_MPI( sym_info, token_value, "MPI_Wait"));
#else
  */
  return (Is_MPI( sym_info, token_value, "MPI_Wait()  "));
/*
  #endif
  */
}


bool_t Is_MPI_Send( SymbolicInfo_t *sym_info,
                     unsigned int token_value )
{
/*
  #ifdef VAMPIR
  return (Is_MPI( sym_info, token_value, "MPI_Send"));
#else
  */
  return (Is_MPI( sym_info, token_value, "MPI_Send() "));
/*
  #endif
  */
}


bool_t Is_MPI_Recv( SymbolicInfo_t *sym_info,
                          unsigned int token_value) /* ALWAYS FUNCTION */
{
/*
  #ifdef VAMPIR
  return (Is_MPI( sym_info, token_value, "MPI_Recv"));
#else
  */
  return (Is_MPI( sym_info, token_value, "MPI_Recv() "));
/*
  #endif
  */
}


void Print_Otf_Symbolic_Info( SymbolicInfo_t *sym_info )
{
  int ii;

  fprintf( stdout, "Otf_Symbolic_Info: num_tokens = %i\n", sym_info->num_tokens );
  for ( ii = 0; ii < sym_info->num_tokens; ii++ )
  {
    Print_Otf_Token( &sym_info->token_info[ ii ], ii );
  }
}


void Free_Otf_Symbolic_Info( SymbolicInfo_t *sym_info )
{
  int ii;

  for ( ii = 0; ii < sym_info->num_tokens; ii++ )
  {
    Free_Otf_Token( &sym_info->token_info[ ii ] );
  }
  free( sym_info );
  sym_info = NULL;
}


static void Partial_Generate_PCF( SymbolicInfo_t *si, FILE *fd, char group, char type )
{
  int ii, jj;
  bool_t print_first_value, values_found;

  for ( ii = 0; ii < Inserted_Tokens( si ); ii++ )
  {
    if ( si->token_info[ ii ].otf_token_type == group )
    {
      fprintf( fd, "\nEVENT_TYPE\n" );
      if ( group != COUNTER_GROUP_TYPE )
        fprintf( fd, "%i\t%i\t%s\n",
                    si->token_info[ ii ].prv_color,
                    si->token_info[ ii ].otf_token_id,
                    si->token_info[ ii ].otf_token_name );
      else
      {
        fprintf( fd, "%i\t%i\t%s\n",
                    si->token_info[ ii ].prv_color,
                    Get_Token_MaskedId( si, si->token_info[ ii ].otf_token_id ),
                    si->token_info[ ii ].otf_token_name );
        
      }

      print_first_value = TRUE;
      for ( jj = 0; jj < Inserted_Tokens( si ); jj++ )
      {
        if (( si->token_info[ jj ].otf_token_type == type ) &&
            ( si->token_info[ jj ].otf_token_group == si->token_info[ ii ].otf_token_id ))
        {
          if ( print_first_value )
          {
            fprintf( fd, "VALUES\n" );
            print_first_value = FALSE;
          }

          fprintf( fd, "%i\t%s\n",
                        si->token_info[ jj ].otf_token_id,
                        si->token_info[ jj ].otf_token_name );
        }
      }
      
      values_found = !print_first_value;
      
      if (( group != COUNTER_GROUP_TYPE ) && 
          ( type  != COUNTER_TYPE ) && 
          ( values_found ))
        fprintf( fd, "0\tEnd\n" );
    }
  }
}


void Generate_PCF( SymbolicInfo_t *si, FILE *fd )
{
  int num_colors;

  /* ASSIGN COLORS */

  /* SECTION: DEFAULT OPTIONS */
  fprintf( fd, "\nDEFAULT_OPTIONS\n" );
  fprintf( fd, "\nLEVEL\t\tTHREAD\n" );
  fprintf( fd, "UNITS\t\tMICROSEC\n" );
  fprintf( fd, "LOOK_BACK\t100\n" );
  fprintf( fd, "SPEED\t\t1\n" );
  fprintf( fd, "FLAG_ICONS\tENABLED\n" );
  num_colors = Inserted_Tokens( si );
  if ( num_colors < DEFAULT_COLORS_PCF )
    num_colors = DEFAULT_COLORS_PCF;
  fprintf( fd, "NUM_OF_STATE_COLORS\t256\n");

  /* SECTION: DEFAULT_SEMANTIC */
  fprintf( fd, "\nDEFAULT_SEMANTIC\n" );
  fprintf( fd, "\nTHREAD_FUNC\tLast Evt Val\n" );

  /* SECTION: EVENTS AND VALUES */
  Partial_Generate_PCF( si, fd, FUNCTION_GROUP_TYPE, FUNCTION_TYPE );
  Partial_Generate_PCF( si, fd, COUNTER_GROUP_TYPE,  COUNTER_TYPE );
}
