/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file   paraver_header.c
 *
 *  @brief  Provides ParaverHeader_t operations.
 *
 *  @author Pedro A. Gonzalez - CEPBA-tools
 */


/* ---[ SYSTEM INCLUDES ]---------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <sys/time.h>
#include <time.h>

/* ---[ OTF2PRV INCLUDES ]--------------------------------------------------- */
#include "paraver_header.h"
#include "log.h"

/* ---[ GLOBAL VARIABLES ]--------------------------------------------------- */
extern ParaverHeader_t *header;
extern FILE *err_fd, *log_fd;

/* ---[ PUBLIC FUNCTIONS ]--------------------------------------------------- */
ParaverHeader_t *Create_Paraver_Header( void )
{
  ParaverHeader_t *header;

  header = (ParaverHeader_t *)malloc( sizeof( ParaverHeader_t ));
  if ( header != NULL )
  {
    header->time_header = NULL;
    strcpy( header->time_units, "us" );
    header->max_time = 0;
    header->process_hier = Create_Process_Hierarchy();
    header->resource_hier = Create_Resource_Hierarchy();
  }
  else
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }


  return header;
}

void Set_Max_Time( ParaverHeader_t *header, uint64_t time )
{
  if ( header != NULL )
    if ( time > header->max_time )
      header->max_time = time;
}

uint64_t Get_Max_Time( ParaverHeader_t *header )
{
  uint64_t ret_time = 0;

  if ( header != NULL )
    ret_time = header->max_time;

  return ( ret_time );
}

void Generate_Paraver_Header( ParaverHeader_t *header,  FILE *fd )
{
  unsigned int node, apps, task;
  struct timeval  tp;
  struct timezone tzp;
  time_t          clock;
  struct tm       *ptm;
  char date[30];

  if ( header != NULL )
  {
    Write_Log( log_fd, SHADOW, "Starting Paraver header generation..." );

    fprintf( fd, "#Paraver ");

    /* translation time */
    gettimeofday (&tp, &tzp);
    clock = tp.tv_sec;
    ptm = localtime (&clock);
    strftime (date, 30, "%d/%m/%y at %H:%M", ptm);

    fprintf( fd, "(%s)", date );
    fprintf( fd, ":" );

    /* total time and units */
    fprintf( fd, "%lld", (long long) header->max_time);
    fprintf( fd, "_%s", (char *)header->time_units );

    fprintf( fd, ":" );

    /* resource hierarchy */
    node = 0;
    fprintf( fd, "%i(%i", header->resource_hier->nnodes,
                          header->resource_hier->node_info[ node ].ncpus );
    ++node;
    while ( node < header->resource_hier->nnodes )
    {
      fprintf( fd, ",%i", header->resource_hier->node_info[ node ].ncpus );
      node++;
    }
    fprintf( fd, "):" );

    /* process_hierarchy */
    fprintf( fd, "%u", header->process_hier->nappls );
    for ( apps = 0; apps < header->process_hier->nappls; apps++)
    {
      fprintf( fd, ":%u(", header->process_hier->appls[ apps ].ntasks );
      for ( task = 0; task < header->process_hier->appls[ apps ].ntasks; task++)
      {
        fprintf( fd, "%u:%u",
                     header->process_hier->appls[ apps ].tasks[ task ].nthreads,
                     header->process_hier->appls[ apps ].tasks[ task ].node);
        if ( task < header->process_hier->appls[ apps ].ntasks - 1 )
          fprintf( fd, "," );
      }
      fprintf( fd, ")" );
    }

    fprintf( fd, "\n" );

    Write_Log( log_fd, SHADOW, "OK\n" );

  }
  else
     Write_Log( err_fd, SHADOW, "Error: Header == NULL!.\n" );


}

void Free_Paraver_Header( ParaverHeader_t *header )
{
  if ( header != NULL )
  {
    if ( header->time_header == NULL )
      free ( header->time_header );
    Free_Process_Hierarchy( header->process_hier );
    Free_Resource_Hierarchy( header->resource_hier );
  }
}
