/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file   paraver_hierarchy.c
 *
 *  @brief  Provides ProcessHierarchy_t and ResourceHierarchy_t operations.
 *
 *  @author Pedro A. Gonzalez - CEPBA-tools
 */

/* ---[ SYSTEM INCLUDES ]--------------------------------------------------- */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
/* ---[ OTF2PRV INCLUDES ]--------------------------------------------------- */
#include "paraver_hierarchy.h"

/* ---[ LOCAL HEADERS ]------------------------------------------------------ */
static void Init_Resource_Hierarchy( ResourceHierarchy_t *res );
static void Init_Process_Hierarchy( ProcessHierarchy_t *ph );

/* ---[ HEADERS ]------------------------------------------------------------ */
/**
 * PROCESS MODEL HIERARCHY
 */

/**
 * Allocates and initializes ParaverHeader_t fields.
 *
 * @param proc_hierarchy  (In) Reference to predefined ProcessHierarchy_t, or
 *                        NULL if none defined.
 * @param res_hierarchy   (In) Reference to predefined ResourceHierarchy_t,
 *                        or NULL if none defined.
 * @return  Initialized ParaverHeader_t structure, or NULL if error.
 */

/**
 *
 * @param header_file_name  (In) Reference to predefined ProcessHierarchy_t, or
 *                          NULL if none defined.
 * @return
 */
static void Init_Process_Hierarchy( ProcessHierarchy_t *ph )
{
  unsigned int ap;
  unsigned int task;
  size_t size;
  
  ph->nappls = 1;
  ph->appls =
    (Application_t *)malloc( sizeof( Application_t ) * MAX_APPLICATIONS );

  if (ph->appls == NULL)
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }

  for ( ap = 0; ap < MAX_APPLICATIONS; ap++)
  {
    ph->appls[ ap ].ntasks = 0;
    ph->appls[ ap ].id_group = 0;
    
    size = (size_t)(sizeof( Task_t ) * MAX_TASKS);
    ph->appls[ ap ].tasks = (Task_t *)malloc( size );
    if ( ph->appls[ ap ].tasks== NULL)
    {
      printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
      exit(1);
    }
    /* memset( ph->appls[ ap ].tasks, '\0', size ); */

    ph->appls[ ap ].ncounters = 0;
    
    size = (size_t)( sizeof( uint64_t ) * MAX_COUNTERS );
    ph->appls[ ap ].counter_index = (uint64_t *)malloc( size );
    if ( ph->appls[ ap ].counter_index == NULL)
    {
      printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
      exit(1);
    }
    /* memset( ph->appls[ ap ].counter_index, '\0', size ); */

    for ( task = 0; task < MAX_TASKS; task++)
    {
      ph->appls[ ap ].tasks[ task ].nthreads = 1;
      ph->appls[ ap ].tasks[ task ].node = task + 1;

      size = (size_t)( sizeof( unsigned int ) * MAX_THREADS );
      ph->appls[ ap ].tasks[ task ].global_cpu = (unsigned int *)malloc( size );

      if ( ph->appls[ ap ].tasks[ task ].global_cpu== NULL)
      {
        printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
        exit(1);
      }

      /* memset( ph->appls[ ap ].tasks[ task ].global_cpu, 0, size ); */
 
      ph->appls[ ap ].tasks[ task ].ncounters = 0;
      size = (size_t)( sizeof( uint64_t ) * MAX_COUNTERS );
      ph->appls[ ap ].tasks[ task ].counter_value = (uint64_t *)malloc( size );
      
      if ( ph->appls[ ap ].tasks[ task ].counter_value == NULL )
      {
        printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
        exit(1);
      }
      /* memset( ph->appls[ ap ].tasks[ task ].counter_value, 0, size ); */
    }
  }
}


/* Inserts that counter for the application */
void InsertCounter( ProcessHierarchy_t *ph, uint32_t counter )
{
  unsigned int task, ap, index;

  ap = 0; /* vampir */
  
  index = ph->appls[ ap ].ncounters;
  ph->appls[ ap ].counter_index[ index ] = counter;
  
  for( task = 0; task < MAX_TASKS; task++ )
  {
    ph->appls[ ap ].tasks[ task ].ncounters++;
    ph->appls[ ap ].tasks[ task ].counter_value[ index ] = 0;
  }

  ph->appls[ ap ].ncounters++;
}

uint64_t GetCounterValue(ProcessHierarchy_t *ph, unsigned int task, uint32_t counter )
{
  unsigned int index = 0;
  
  for ( index = 0; index < ph->appls[ 0 ].ncounters; index++ )
  {
    if ( counter ==  ph->appls[ 0 ].counter_index[ index ] )
      break;
  }
    
  return ph->appls[ 0 ].tasks[ task ].counter_value[ index ];
}

void SetCounterValue(ProcessHierarchy_t *ph, unsigned int task, uint32_t counter, uint64_t value )
{
  unsigned int index = 0;
  
  for ( index = 0; index < ph->appls[ 0 ].ncounters; index++ )
  {
    if ( counter ==  ph->appls[ 0 ].counter_index[ index ] )
      break;
  }
    
  ph->appls[ 0 ].tasks[ task ].counter_value[ index ] = value;
}

ProcessHierarchy_t *Create_Process_Hierarchy( void )
{
  ProcessHierarchy_t *ph;

  ph = (ProcessHierarchy_t *)malloc( sizeof( ProcessHierarchy_t ));
  if ( ph != NULL )
    Init_Process_Hierarchy( ph );
  else
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }

  return ph;
}

void Print_Process_Hierarchy( ProcessHierarchy_t *ph )
{
  unsigned int ap;
  unsigned int task;

  if ( ph == NULL )
    printf("WARNING: Void process hierarchy!\n");
  else
  {
    printf( "Process Hierarchy: %i applications.\n", ph->nappls );

    for ( ap = 0; ap < ph->nappls; ap++)
    {
      printf("\tApp[\t%i] with %i tasks and %i id_group.\n", 
        ap,
        ph->appls[ ap ].ntasks,
        ph->appls[ ap ].id_group );

      for ( task = 0; task < ph->appls[ ap ].ntasks; task++)
      {
        printf("\t\tTask[\t%i] with %i threads executing in node %i.\n",
          task,
          ph->appls[ ap ].tasks[ task ].nthreads,
          ph->appls[ ap ].tasks[ task ].node );
      }
    }
  }
}


void Free_Process_Hierarchy( ProcessHierarchy_t *ph )
{
  unsigned int ap;
  unsigned int task;

  if ( ph != NULL )
  {
    for ( ap = 0; ap < MAX_APPLICATIONS; ap++)
    {
      for ( task = 0; task < MAX_TASKS; task++)
      {
        free( ph->appls[ ap ].tasks[ task ].global_cpu );
        free( ph->appls[ ap ].tasks[ task ].counter_value );
      }
      free( ph->appls[ ap ].tasks );
      free( ph->appls[ ap ].counter_index );
    }
    free( ph->appls );
  }
  free( ph );
}

/**
 * RESOURCE MODEL HIERARCHY
 */

/**
 * Creates a file named 'header_file_name', containing header information
 * gathered from ParaverHeader_t fields.
 *
 * @param header_file_name  (In) Reference to predefined ProcessHierarchy_t, or
 *                          NULL if none defined.
 * @return  Header file, or NULL if error.
 */
static void Init_Resource_Hierarchy( ResourceHierarchy_t *res )
{
  int node;
  unsigned int gcpu;

  res->nnodes = 0;
  res->node_info = (NodeInfo_t *)malloc( sizeof( NodeInfo_t ) * MAX_NODES );
if ( res->node_info== NULL)
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }

  gcpu = 1;
  for ( node = 0; node < MAX_NODES; node++)
  {
    res->node_info[ node ].ncpus = DEFAULT_CPUS_PER_NODE;
    res->node_info[ node ].global_cpu = gcpu++;
  }
}

ResourceHierarchy_t *Create_Resource_Hierarchy( void )
{
  ResourceHierarchy_t *res;

  res = (ResourceHierarchy_t *)malloc( sizeof( ResourceHierarchy_t ));
  if ( res != NULL )
    Init_Resource_Hierarchy( res );
  else
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }

  return res;
}


/**
 * Prints a resource hierarchy.
 *
 * @param header_file_name  (In) Reference to predefined ReferenceHierarchy_t, or
 *                          NULL if none defined.
 */
void Print_Resource_Hierarchy( ResourceHierarchy_t *res )
{
  int node;

  if ( res == NULL )
    printf("WARNING: Void resource hierarchy!\n");
  else
  {
    printf( "Resouce Hierarchy: %i nodes\n", res->nnodes );
    for ( node = 0; node < res->nnodes; node++)
    {
      printf( "\tNode[\t%i] has %i cpus with global cpu %i.\n",
        node,
        res->node_info[ node ].ncpus,
        res->node_info[ node ].global_cpu );
    }
  }
}


void Free_Resource_Hierarchy( ResourceHierarchy_t *res )
{
  if ( res != NULL )
  {
    if ( res->nnodes != 0 )
      free( res->node_info );

    free( res );
  }
}
