/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file  parse_parameters.c
 *
 *  @brief  Functions related with options and parameters.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "parse_parameters.h"


void Show_Usage( void )
{
  fprintf( stdout, "%s%s%s", USAGE_MSG, OPTIONS_MSG, PARAMETERS_MSG );
}


TranslatorOptions_t *Create_Options( void )
{
  TranslatorOptions_t *opt;

  opt = (TranslatorOptions_t *)malloc( sizeof(TranslatorOptions_t));
  if (opt == NULL)
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }

  opt->prv_path       = NULL;
  opt->prv_trace_name = NULL;
  opt->otf_path       = NULL;
  opt->otf_trace_name = NULL;
  opt->events_path       = NULL;
  opt->events_trace_name = NULL;
  bzero( opt->option, (size_t)(sizeof(bool_t)*(NUM_OPTIONS+1)) );

  return opt;
}


void Print_Options( TranslatorOptions_t *to )
{
  int ii;

  if ( to != NULL )
  {
    if ( to->prv_path != NULL )
      printf( "Paraver trace path: %s\n", to->prv_path );

    if ( to->prv_trace_name != NULL )
      printf( "Paraver trace name: %s\n", to->prv_trace_name );

    if ( to->otf_path != NULL )
      printf( "Otf trace path: %s\n", to->otf_path );

    if ( to->otf_trace_name != NULL )
      printf( "Otf trace name: %s\n", to->otf_trace_name );

    if ( to->events_path != NULL )
      printf( "Events trace path: %s\n", to->events_path );

    if ( to->events_trace_name != NULL )
      printf( "Events trace name: %s\n", to->events_trace_name );

    for ( ii = 1; ii <= NUM_OPTIONS; ii ++)
      printf( "Options: %i\n", to->option[ii] );
  }
}

void Free_Options( TranslatorOptions_t *to )
{
  free( to->prv_path );
  to->prv_path = NULL;
  free( to->prv_trace_name );
  to->prv_trace_name = NULL;
  free( to->otf_path );
  to->otf_path = NULL;
  free( to->otf_trace_name );
  to->otf_trace_name = NULL;
  free( to->events_path );
  to->events_path = NULL;
  free( to->events_trace_name );
  to->events_trace_name = NULL;
  free( to );
}

char *Get_EVENTS_Name( TranslatorOptions_t *to )
{
  char *aux_name, *evt_name;
  int length;

  if ( to->events_trace_name == NULL )
  {
    length = strlen( to->otf_trace_name ) + 5;
    evt_name = (char *)malloc(  sizeof(char) * length );
    
    if (evt_name == NULL)
    {
      printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
      exit(1);
    }
    strncpy( evt_name, to->otf_trace_name, strlen(to->otf_trace_name) );
    aux_name = evt_name + strlen( evt_name ) - 3;
    strcpy( aux_name, "1.events" );
    to->events_trace_name = evt_name;
  }

  return to->events_trace_name;
}

char *New_File_Name( TranslatorOptions_t *to, const char* suffix )
{
  char *aux_otf, *tmp_name;

  if ( to->prv_trace_name == NULL )
    tmp_name = strdup( to->otf_trace_name );
  else
    tmp_name = strdup( to->prv_trace_name );

  aux_otf  = tmp_name + strlen( tmp_name ) - strlen( suffix );
  strcpy( aux_otf, suffix );

  return tmp_name;
}


void Parse_Options( int argc, char *argv[], TranslatorOptions_t *param )
{
  int ii;

  for ( ii=1; ii<argc; ii++ )
  {
    if (( strcmp( argv[ii], "-h" ) == 0 ) ||
        ( strcmp( argv[ii], "--help" ) == 0 ))
    {
      param->option[ OPTION_HELP ] = TRUE;
    }
    else if (( strcmp( argv[ii], "-l" ) == 0 ) ||
        ( strcmp( argv[ii], "--log" ) == 0 ))
    {
      param->option[ OPTION_LOG ] = TRUE;
    }
    else if (( strcmp( argv[ii], "-v" ) == 0 ) ||
        ( strcmp( argv[ii], "--verbose" ) == 0 ))
    {
      param->option[ OPTION_VERBOSE ] = TRUE;
    }
    else if (( strcmp( argv[ii], "-i2" ) == 0 ) ||
        ( strcmp( argv[ii], "--ignore-events" ) == 0 ))
    {
      param->option[ OPTION_I2 ] = TRUE;
    }
    else if (( strcmp( argv[ii], "-i3" ) == 0 ) ||
        ( strcmp( argv[ii], "--ignore-comms" ) == 0 ))
    {
      param->option[ OPTION_I3 ] = TRUE;
    }
    else if ( strcmp( argv[ii] + strlen( argv[ii ]) - 4, SUFFIX_DOT_PRV ) == 0 )
    {
      param->option[ OPTION_PRV_GZIPPED ] = FALSE;
      param->prv_trace_name = strdup( argv[ii] );
    }
    else if ( strcmp( argv[ii] + strlen( argv[ii ]) - 7, SUFFIX_DOT_PRV_GZ ) == 0 )
    {
      param->option[ OPTION_PRV_GZIPPED ] = TRUE;
      param->prv_trace_name = strdup( argv[ii] );
    }
    else if ( strcmp( argv[ii] + strlen( argv[ii ]) - 4, SUFFIX_DOT_OTF ) == 0 )
    {
      param->option[ OPTION_OTF ] = TRUE;
      param->option[ OPTION_OTF_GZIPPED ] = FALSE;
      param->otf_trace_name = strdup( argv[ii] );
    }
  }
}
