/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file  read_handlers.c
 *
 *  @brief  Defines handlers behaviour.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tgmath.h>

#include "otf.h"
#include "otf_symbolic_info.h"
#include "read_handlers.h"
#include "types.h"
#include "comms_queue.h"
#include "paraver_header.h"
#include "log.h"
#include "statistics.h"
#include "paraver_hierarchy.h"

extern FILE *tmp_fd, *pcf_fd, *err_fd, *log_fd, *comm_fd;
extern ParaverHeader_t *header;
extern TranslatorOptions_t *parameters;

int queued = 0;
int dequeued = 0;

extern EntryStack_t  *function_stack;

#ifdef VAMPIR
extern EntryStack_t  *enter_events_stack;
extern uint64_t ticks_per_second;
extern uint64_t unit_factor;
#endif

extern ThreadCommsQueue_t *comms;
extern ThreadCommsQueue_t *unmatched_receives;
extern Statistics_t stats;

EventRecord_t last_printed = {0,0,0,0,0,0,0};

unsigned char current_profile = DEFAULT_PROFILE;

Translator_Profiles_t Profile_Settings[] =
{
  { "none",    "DEFAULT", "MICROSEC", "us", 6,    1 },
  { "tau2otf", "TAU",     "MICROSEC", "us", 6, 1000 },
  { "vampir",  "VAMPIR",  "MICROSEC", "us", 6, 1000 },
};

/* LOCAL HEADERS */
static void Print_EventRecord( EventRecord_t *er, FILE *fd );
static void Print_EventRecord_Ext( FILE *fd,
                                    uint32_t cpu,
                                    uint32_t ptask,
                                    uint32_t task,
                                    uint32_t thread,
                                    uint64_t time,
                                    uint32_t type,
                                    uint32_t value );
static void Print_CommunicationRecord( CommRecord_t *cr, FILE *fd );

static void Event_Detected( uint64_t time,
                            uint32_t function,
                            uint32_t process,
                            uint32_t source,
                            bool_t   exit );
                           
static void Counter_Event_Detected( uint64_t time,
                                    uint32_t process,
                                    uint32_t counter,
                                    uint32_t value );
                                    
static void Send_Detected( uint64_t        time,
                           uint32_t        sender,
                           uint32_t        receiver,
                           uint32_t        group,
                           uint32_t        type,
                           uint32_t        length,
                           uint32_t        source );
                           
static void Receive_Detected( uint64_t        time,
                              uint32_t        recvProc,
                              uint32_t        sendProc,
                              uint32_t        group,
                              uint32_t        type,
                              uint32_t        length,
                              uint32_t        source );
                              

/* LOCAL FUNCTIONS */
static uint64_t TimerResolution( uint64_t time )
{
#ifdef VAMPIR
  time = (uint64_t)((double)time * ( (double)unit_factor / (double)ticks_per_second ));
#else
  time = time / (uint64_t)Profile_Settings[ current_profile ].conversion_factor;
#endif

  return time;
}


static void Print_EventRecord_Ext( FILE *fd,
                                    uint32_t cpu,
                                    uint32_t ptask,
                                    uint32_t task,
                                    uint32_t thread,
                                    uint64_t time,
                                    uint32_t type,
                                    uint32_t value )
{
  if ( stats.prv_records[ PRV_EVENTS ] == 0 )
	{
		fprintf( fd, FORMAT_EVENT_RECORD,
                 cpu, ptask, task, thread,
                 (unsigned long long int)time, type,
                 (unsigned long long int)value );
	}
  else if (( cpu    == last_printed.cpu )    &&
           ( ptask  == last_printed.ptask )  &&
				   ( task   == last_printed.task )   &&
           ( thread == last_printed.thread ) &&
           ( time   == last_printed.time ))
  {
    fprintf( fd, FORMAT_EVENT_RECORD_COMPACT,
                 type, (unsigned long long int)value );
  }
  else
  {
    fprintf( fd, "\n" );
    fprintf( fd, FORMAT_EVENT_RECORD,
                 cpu, ptask, task, thread,
                 (unsigned long long int)time, type,
                 (unsigned long long int)value );
  }

  Set_Max_Time( header, time );
  stats.prv_records[ PRV_EVENTS ]++;

  last_printed.cpu    = cpu;
  last_printed.ptask  = ptask;
  last_printed.task   = task;
  last_printed.thread = thread;
  last_printed.time   = time; 
}


static void Print_EventRecord( EventRecord_t *er, FILE *fd )
{
  Print_EventRecord_Ext( fd,
                         er->cpu,
                         er->ptask,
                         er->task,
                         er->thread,
                         er->time,
                         er->type,
                         er->value );
}


static void Print_CommunicationRecord( CommRecord_t *cr, FILE *fd )
{
  fprintf( fd, FORMAT_COMMS_RECORD,
           cr->snd_cpu, cr->snd_ptask, cr->snd_task, cr->snd_thread,
           (unsigned long long int)cr->snd_time, (unsigned long long int)cr->snd_time,
           cr->rcv_cpu, cr->rcv_ptask, cr->rcv_task, cr->rcv_thread,
           (unsigned long long int)cr->rcv_time, (unsigned long long int)cr->rcv_time,
           cr->size, cr->tag );

  stats.prv_records[ PRV_COMMUNICATIONS ]++;
  if ( cr->snd_time > cr->rcv_time )
  {
    stats.prv_negative_comms++;
    Set_Max_Time( header, cr->snd_time );
  }
  else
    Set_Max_Time( header, cr->rcv_time );
}

static void Event_Detected( uint64_t time,
                            uint32_t function,
                            uint32_t process,
                            uint32_t source,
                            bool_t   exit )
{
  int cpu, ptask, task, thread;
  unsigned int tag;
  EventRecord_t *er;
  EventRecord_t *exit_event;

  cpu = 0;
  ptask = thread = 1;
  task = process;
  time = TimerResolution( time );
	
  if ( !Get_Otf_Token_Group( sym_otf_info, function, FUNCTION_TYPE , &tag ))
	  return;
  
	Set_Max_Time( header, time );

  if ( exit )
  {
    Print_EventRecord_Ext( tmp_fd, process, ptask, process, thread,
                           ( unsigned long long int )time, tag,
                           ( unsigned long long int )0);
    if (!IsEmptyEventsQueue( function_stack->entry[process] ))
    {
      GetFirstEvent( function_stack->entry[process], &exit_event );
      if ( exit_event->value == function )
      {
        DequeueFirstEvent(function_stack->entry[process]);
        FreeEventRecord( exit_event );
      }
    }
  }
  else /* entry */
  {
    Print_EventRecord_Ext( tmp_fd, process, ptask, process, thread,
                           ( unsigned long long int )time, tag,
                           ( unsigned long long int )function);

    er = CreateEventRecord( process, ptask, process, thread, time, tag, function );
    PushEventStack( function_stack, er, process, TRUE );
    /* FreeEventRecord(er); */
  }
}


static void Counter_Event_Detected( uint64_t time,
                                    uint32_t process,
                                    uint32_t counter,
                                    uint32_t value )
{
  int cpu, ptask, task, thread;
  /* unsigned int tag; */
#ifdef VAMPIR
  uint64_t prev_value, differential = 0;
  unsigned int masked_id;
#endif
    
  cpu = 0;
  ptask = thread = 1;

  task = process;

  time = TimerResolution( time );
  /* tag = Get_Otf_Token_Group( sym_otf_info, function, COUNTER_GROUP_TYPE );*/
  Set_Max_Time( header, time );
  /* PENDIENTE MASKED_ID */
  counter = Get_Otf_Token_AliasId( sym_otf_info, counter, COUNTER_GROUP_TYPE, &masked_id );

#ifdef VAMPIR
  /* Compute differential value */
  counter = Get_Token_MaskedId( sym_otf_info, counter );

  prev_value = GetCounterValue( header->process_hier, task, counter );
  
  if ( prev_value != 0 )
    differential = value - prev_value;
  
  SetCounterValue( header->process_hier, task, counter, value );

  value = differential;
#endif
  
  Print_EventRecord_Ext( tmp_fd, process, ptask, process, thread,
                         (unsigned long long int)time, counter, 
                         (unsigned long long int)value);
}


/* Performs dump of matched communication */
void LimitedDump_Comms( void )
{
  bool_t finished;
  CommRecord_t *cm;

  finished = FALSE;
  cm = FirstGlobalComm( comms );
  
	
  while( ( cm != NULL ) && !finished )
  {
    if ( cm->matched )
    {
      Print_CommunicationRecord( cm, comm_fd );
      DequeueComm( comms, cm->snd_cpu );
      stats.queue[ QSENDS ].dequeued++;
      stats.queue[ QSENDS ].total_sum_queue += comms->num_comms;

      cm = FirstGlobalComm( comms );
    }
    else
      finished = TRUE;
  }
}


void Dump_Communications( ThreadCommsQueue_t *cq )
{
  bool_t finish;
  CommRecord_t *cm;

  finish = FALSE;
  do
  {
    cm = FirstGlobalComm( cq );
    if ( cm == NULL)
      finish =  TRUE;
    else 
    {
      fprintf(
        comm_fd,
        FORMAT_COMMS_RECORD,
        cm->snd_cpu, cm->snd_ptask, cm->snd_task, cm->snd_thread,
        (unsigned long long int)cm->snd_time, (unsigned long long int)cm->snd_time,
        cm->rcv_cpu, cm->rcv_ptask, cm->rcv_task, cm->rcv_thread,
        (unsigned long long int)cm->rcv_time, (unsigned long long int)cm->rcv_time,
        cm->size, cm->tag );
      
      DequeueComm( cq, cm->snd_cpu );
      stats.queue[ QSENDS ].dequeued++;
      stats.queue[ QSENDS ].total_sum_queue += comms->num_comms;
    }
  }while (!finish);
}

static void Send_Detected( uint64_t        time,
                           uint32_t        sender,
                           uint32_t        receiver,
                           uint32_t        group,
                           uint32_t        type,
                           uint32_t        length,
                           uint32_t        source )
{
  int ptask, thread;
  CommRecord_t *cr;
  CommRecord_t *pending;
/*  
  EventRecord_t *last_evt = NULL;
*/
  uint32_t mpitype = MPI_UNKNOWN;

  ptask = thread = 1;
  time = TimerResolution( time );
  Set_Max_Time( header, time );

/*
  if ( !IsEmptyEventsQueue( enter_events_stack->entry[ sender ] ))
  {
    GetFirstEvent( enter_events_stack->entry[ sender ], &last_evt );
    if ( Is_MPI_Send( sym_otf_info, last_evt->value ))
      mpitype = MPI_SEND;
  }
*/  
  cr = CreateCommRecord( sender, ptask, sender, thread, time,
                         receiver, ptask, receiver, thread, 0,
                         length, type, mpitype );

  pending = SearchComm( unmatched_receives, sender, cr );
  if ( pending != NULL )
  {
    cr->rcv_time = pending->rcv_time;
    cr->matched  = TRUE;
  
    pending->snd_time = cr->snd_time;
    pending->matched  = TRUE;

    stats.otf_matched_sends++;
	}

  QueueComm( comms, sender, cr );
  stats.queue[ QSENDS ].queued++;
  stats.queue[ QSENDS ].total_sum_queue+= comms->num_comms;
}


static void Receive_Detected( uint64_t        time,
                              uint32_t        recvProc,
                              uint32_t        sendProc,
                              uint32_t        group,
                              uint32_t        type,
                              uint32_t        length,
                              uint32_t        source )
{
  int ptask, thread;
  CommRecord_t *rcv;
  CommRecord_t *pending = NULL;
/*
  EventRecord_t *last_evt = NULL;
*/
  uint32_t mpitype = MPI_UNKNOWN;

  ptask = thread = 1;
  time = TimerResolution( time );
  Set_Max_Time( header, time );

/*
  if ( !IsEmptyEventsQueue( enter_events_stack->entry[ recvProc ] ))
  {
    GetFirstEvent( enter_events_stack->entry[ recvProc ], &last_evt );
    if ( Is_MPI_Wait( sym_otf_info, last_evt->value ))
      mpitype = MPI_WAIT;
    else if (Is_MPI_Recv(sym_otf_info, last_evt->value))
      mpitype = MPI_RECV;
  }
*/
  
  rcv = CreateCommRecord( sendProc, ptask, sendProc, thread, 0,
                          recvProc, ptask, recvProc, thread, time,
                          length, type, mpitype );
  
  pending = SearchComm( comms, sendProc, rcv );
  if ( pending != NULL )
  {
    pending->rcv_time = rcv->rcv_time;
    pending->matched  = TRUE;
    
    stats.otf_matched_receives++;
    FreeCommRecord( rcv );

    dequeued++;
		
		LimitedDump_Comms( );
  }
  else
  {
    QueueComm( unmatched_receives, sendProc, rcv );
    stats.queue[ QRECEIVES ].queued++;
    stats.queue[ QRECEIVES ].total_sum_queue += unmatched_receives->num_comms;
  }
}


int ReadHandler_DefinitionComment(  void*           userData,
                                    uint32_t        stream,
                                    const char*     comment )
{
  stats.otf_records[ OTF_DEFCOMMENT_REC ]++;

  return 0;
}

int ReadHandler_DefTimerResolution( void*           userData,
                                    uint32_t        stream,
                                    uint64_t        ticksPerSecond )
{
  int exp;

  stats.otf_records[ OTF_DEFTIMERRESOLUTION_REC ]++;

  if ( current_profile == TAU_PROFILE )
  {
     exp = Profile_Settings[ current_profile ].exp;
     strncpy( header->time_units,
              Profile_Settings[ current_profile ].units_name_short,
              strlen(Profile_Settings[ current_profile ].units_name_short));
  }
  else
  {
    ticks_per_second = ticksPerSecond;
    
    exp = (int)floor(log10(ticksPerSecond));
    switch(exp)
    {
      case 0:
        strncpy( header->time_units, "s", strlen("s"));
        unit_factor = 1;
        break;
      case 3:
        strncpy( header->time_units, "ms", strlen("ms"));
        unit_factor = 1000;
        break;
      case 6:
        strncpy( header->time_units, "us", strlen("us"));
        unit_factor = 1000000;
        break;
      case 9:
        strncpy( header->time_units, "ns", strlen("ns"));
        unit_factor = 1000000000;
        break;
      default:
        unit_factor = 1000000;
        strncpy( header->time_units,
                 Profile_Settings[ DEFAULT_PROFILE ].units_name_short,
                 strlen(Profile_Settings[ DEFAULT_PROFILE ].units_name_short));
        break;
    }
  }

  return 0;
}


int ReadHandler_DefProcess(         void*           userData,
                                    uint32_t        stream,
                                    uint32_t        process,
                                    const char*     name,
                                    uint32_t        parent )
{
  unsigned int index, i_th, app;

  stats.otf_records[ OTF_DEFPROCESS_REC ]++;

  header->resource_hier->nnodes++;
  app =  header->process_hier->nappls-1;
  header->process_hier->appls[ app ].ntasks++;

/*
  if ( header->process_hier->appls[ app ].ntasks > 0 )
  {
*/
    index = header->process_hier->appls[ app ].ntasks - 1;


    i_th = header->process_hier->appls[ app ].tasks[ index ].nthreads-1;

    header->process_hier->appls[ app ].tasks[ index ].global_cpu[ i_th ] = 1;
    header->process_hier->appls[ app ].tasks[ index ].node =
      header->resource_hier->nnodes;
    header->process_hier->appls[ app ].tasks[ index ].nthreads++;
/*
  }
*/
  return 0;
}

int ReadHandler_DefProcessGroup(    void*           userData,
                                    uint32_t        stream,
                                    uint32_t        procGroup,
                                    const char*     name,
                                    uint32_t        numberOfProcs,
                                    const uint32_t* procs )
{
/*
  unsigned int app;
*/
  stats.otf_records[ OTF_DEFPROCESSGROUP_REC ]++;

/*
  app = header->process_hier->nappls;
  header->process_hier->appls[ app ].ntasks = numberOfProcs;
  header->process_hier->appls[ app ].id_group = procGroup;
  header->process_hier->nappls++; */

  return 0;
}

int ReadHandler_DefFunction(        void*           userData,
                                    uint32_t        stream,
                                    uint32_t        func,
                                    const char*     name,
                                    uint32_t        funcGroup,
                                    uint32_t        source )
{
  stats.otf_records[ OTF_DEFFUNCTION_REC ]++;
  Insert_Otf_Info( FUNCTION_TYPE, funcGroup, func, name, sym_otf_info, 0 );

  return 0;
}

int ReadHandler_DefFunctionGroup(   void*           userData,
                                    uint32_t        stream,
                                    uint32_t        funcGroup,
                                    const char*     name )
{
  stats.otf_records[ OTF_DEFFUNCTIONGROUP_REC ]++;
  Insert_Otf_Info( FUNCTION_GROUP_TYPE, 0, funcGroup, (const char *)name, sym_otf_info, 0 );

  return 0;
}

int ReadHandler_DefCollectiveOperation(
                                    void*    userData,
                                    uint32_t        stream,
                                    uint32_t        collOp,
                                    const char*     name,
                                    uint32_t        type )
{
  stats.otf_records[ OTF_DEFCOLLOP_REC ]++;

  return 0;
}

int ReadHandler_DefCounter(         void*           userData,
                                    uint32_t        stream,
                                    uint32_t        counter,
                                    const char*     name,
                                    uint32_t        properties,
                                    uint32_t        counterGroup,
                                    const char*     unit )
{
#ifdef VAMPIR  
  unsigned int masked_id = 0;
#endif  
  
  stats.otf_records[ OTF_DEFCOUNTER_REC ]++;

#ifdef VAMPIR
  /* counter = ( counterGroup << 16 ) | counter; */
  Insert_Otf_Info( COUNTER_GROUP_TYPE, 0, counter, (const char *)name, sym_otf_info, masked_id );
  
    /* PENDIENTE MASKED_ID */
  counter = Get_Otf_Token_AliasId( sym_otf_info, counter, COUNTER_GROUP_TYPE, &masked_id );
  masked_id = ( counterGroup * 1000000 ) + counter;
  
  InsertCounter( header->process_hier, masked_id  );
  Set_Token_MaskedId( sym_otf_info, counter, masked_id );
#else
  Insert_Otf_Info( COUNTER_TYPE, counterGroup, counter, (const char *)name, sym_otf_info, 0 );
#endif
  
  return 0;
}

int ReadHandler_DefCounterGroup(    void*           userData,
                                    uint32_t        stream,
                                    uint32_t        counterGroup,
                                    const char*     name )
{
  stats.otf_records[ OTF_DEFCOUNTERGROUP_REC ]++;

#ifndef VAMPIR
  Insert_Otf_Info( COUNTER_GROUP_TYPE, 0, counterGroup, (const char *)name, sym_otf_info, 0 );
#endif

  return 0;
}

int ReadHandler_DefScl(             void*           userData,
                                    uint32_t        stream,
                                    uint32_t        source,
                                    uint32_t        sourceFile,
                                    uint32_t        line )
{
  stats.otf_records[ OTF_DEFSCL_REC ]++;

  return 0;
}

int ReadHandler_DefSclFile(         void*           userData,
                                    uint32_t        stream,
                                    uint32_t        sourceFile,
                                    const char*     name )
{
  stats.otf_records[ OTF_DEFSCLFILE_REC ]++;

  return 0;
}

int ReadHandler_DefCreator(         void*           userData,
                                    uint32_t        stream,
                                    const char*     creator )
{
  char *creator_msg;
  char *creator_name;
  char *trash;
  int found, ii;

  stats.otf_records[ OTF_DEFCREATOR_REC ]++;

  creator_msg = strdup( creator );
  creator_name = (char *)malloc( sizeof( char ) * strlen( creator_msg ) );
  if (creator_name == NULL)
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }
  trash        = (char *)malloc( sizeof( char ) * strlen( creator_msg ) );
  if (trash == NULL)
  {
    printf("Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__);
    exit(1);
  }
  found = sscanf( creator_msg, "%s %s", creator_name, trash);
  if ( found > 0 )
  {
    for ( ii = 0; ii < MAX_PROFILE; ii ++ )
    {
      if ( !strcmp( creator_name, Profile_Settings[ ii ].creator ) )
        break;
    }
    current_profile = ii;
  }
  else
    current_profile = DEFAULT_PROFILE;

  free( creator_msg );
  free( trash );
  free( creator_name );

  return 0;
}


int ReadHandler_Enter(              void*           userData,
                                    uint64_t        time,
                                    uint32_t        function,
                                    uint32_t        process,
                                    uint32_t        source )
{
#ifdef VAMPIR
  int ptask = 1;
  int task;
  int thread = 1;
  unsigned int tag;
  EventRecord_t *er;
  uint64_t original_time;
#endif
  
  stats.otf_records[ OTF_ENTER_REC ]++;

#ifdef VAMPIR
  task = process;
  
  original_time = time;
  time = TimerResolution( time );
	
  if ( Get_Otf_Token_Group( sym_otf_info, function, FUNCTION_TYPE , &tag))
  {
    er = CreateEventRecord( process, ptask, process, thread, time, tag, function );
  	PushEventStack( enter_events_stack, er, process, TRUE );
    /*FreeEventRecord( er );*/
  }
#endif						
  time = original_time;
  Event_Detected( time, function, process, source, FALSE );

  return 0;
}

int ReadHandler_Leave(              void*           userData,
                                    uint64_t        time,
                                    uint32_t        function,
                                    uint32_t        process,
                                    uint32_t        source )
{
#ifdef VAMPIR
  uint32_t original_function;
	EventRecord_t *last_enter_event;
#endif
  
  stats.otf_records[ OTF_LEAVE_REC ]++;

#ifdef VAMPIR
  original_function = function;
  if ( function == 0 )
  {
	  last_enter_event = NULL;
    if ( !IsEmptyEventsQueue(enter_events_stack->entry[process] ))
    {
      GetFirstEvent( enter_events_stack->entry[process], &last_enter_event );
      DequeueFirstEvent( enter_events_stack->entry[process] );
      function = last_enter_event->value;
      FreeEventRecord( last_enter_event );
    }
  }		
#endif

  Event_Detected( time, function, process, source, TRUE );
  return 0;
}

int ReadHandler_SendMsg(            void*           userData,
                                    uint64_t        time,
                                    uint32_t        sender,
                                    uint32_t        receiver,
                                    uint32_t        group,
                                    uint32_t        type,
                                    uint32_t        length,
                                    uint32_t        source )
{
  stats.otf_records[ OTF_SEND_REC ]++;
  Send_Detected( time, sender, receiver, group, type, length, source );

  return 0;
}

int ReadHandler_RecvMsg(            void*           userData,
                                    uint64_t        time,
                                    uint32_t        recvProc,
                                    uint32_t        sendProc,
                                    uint32_t        group,
                                    uint32_t        type,
                                    uint32_t        length,
                                    uint32_t        source )
{
  stats.otf_records[ OTF_RECEIVE_REC ]++;
  Receive_Detected( time, recvProc, sendProc, group, type, length, source );

  return 0;
}

int ReadHandler_Counter(            void*           userData,
                                    uint64_t        time,
                                    uint32_t        process,
                                    uint32_t        counter,
                                    uint64_t        value )
{
  int cpu, ptask, task, thread;

  stats.otf_records[ OTF_COUNTER_REC ]++;

  cpu = 0;
  ptask = thread = 1;

  /* PROCESS -> cpu:ptask:task:thread */
  task = process;
  Counter_Event_Detected( time, process, counter, value );

  return 0;
}

int ReadHandler_CollectiveOperation(void*           userData,
                                    uint64_t        time,
                                    uint32_t        process,
                                    uint32_t        collective,
                                    uint32_t        procGroup,
                                    uint32_t        rootProc,
                                    uint32_t        sent,
                                    uint32_t        received,
                                    uint64_t        duration,
                                    uint32_t        source )
{
  stats.otf_records[ OTF_COLLOP_REC ]++;

  return 0;
}

int ReadHandler_EventComment(       void*           userData,
                                    uint64_t        time,
                                    uint32_t        process,
                                    const char*     comment )
{
  stats.otf_records[ OTF_EVENTCOMMENT_REC ]++;

  return 0;
}

int ReadHandler_BeginProcess(       void*           userData,
                                    uint64_t        time,
                                    uint32_t        process )
{
  stats.otf_records[ OTF_BEGINPROCESS_REC ]++;

  return 0;
}

int ReadHandler_EndProcess(         void*           userData,
                                    uint64_t        time,
                                    uint32_t        process )
{
  stats.otf_records[ OTF_ENDPROCESS_REC ]++;

  return 0;
}

int ReadHandler_SnapshotComment(    void*           userData,
                                    uint64_t        time,
                                    uint32_t        process,
                                    const char*     comment )
{
  stats.otf_records[ OTF_SNAPSHOTCOMMENT_REC ]++;

  return 0;
}

int ReadHandler_EnterSnapshot(      void*           userData,
                                    uint64_t        time,
                                    uint64_t        originaltime,
                                    uint32_t        function,
                                    uint32_t        process,
                                    uint32_t        source )
{
  stats.otf_records[ OTF_ENTERSNAPSHOT_REC ]++;
  
  return 0;
}

int ReadHandler_SendSnapshot(       void*           userData,
                                    uint64_t        time,
                                    uint64_t        originaltime,
                                    uint32_t        sender,
                                    uint32_t        receiver,
                                    uint32_t        procGroup,
                                    uint32_t        tag,
                                    uint32_t        source )
{
  stats.otf_records[ OTF_SENDSNAPSHOT_REC ]++;

  return 0;
}

int ReadHandler_SummaryComment(     void*           userData,
                                    uint64_t        time,
                                    uint32_t        process,
                                    const char*     comment )
{
  stats.otf_records[ OTF_SUMMARYCONTENT_REC ]++;

  return 0;
}

int ReadHandler_FunctionSummary(    void*           userData,
                                    uint64_t        time,
                                    uint32_t        function,
                                    uint32_t        process,
                                    uint64_t        invocations,
                                    uint64_t        exclTime,
                                    uint64_t        inclTime )
{
  stats.otf_records[ OTF_FUNCTIONSUMMARY_REC ]++;

  return 0;
}

int ReadHandler_FunctionGroupSummary( void*         userData,
                                      uint64_t      time,
                                      uint32_t      funcGroup,
                                      uint32_t      process,
                                      uint64_t      invocations,
                                      uint64_t      exclTime,
                                      uint64_t      inclTime )
{
  stats.otf_records[ OTF_FUNCTIONGROUPSUMMARY_REC ]++;

  return 0;
}

int ReadHandler_MessageSummary(     void*           userData,
                                    uint64_t        time,
                                    uint32_t        process,
                                    uint32_t        peer,
                                    uint32_t        comm,
                                    uint32_t        type,
                                    uint64_t        sentNumber,
                                    uint64_t        receivedNumber,
                                    uint64_t        sentBytes,
                                    uint64_t        receivedBytes )
{
  stats.otf_records[ OTF_MESSAGESUMMARY_REC ]++;

  return 0;
}
