/*****************************************************************************\
 *                        ANALYSIS PERFORMANCE TOOLS                         *
 *                                   otf2prv                                 *
 *                    Trace translator from otf to prv format                *
 *****************************************************************************
 *     ___     This library is free software; you can redistribute it and/or *
 *    /  __         modify it under the terms of the GNU LGPL as published   *
 *   /  /  _____    by the Free Software Foundation; either version 2.1      *
 *  /  /  /     \   of the License, or (at your option) any later version.   *
 * (  (  ( B S C )                                                           *
 *  \  \  \_____/   This library is distributed in hope that it will be      *
 *   \  \__         useful but WITHOUT ANY WARRANTY; without even the        *
 *    \___          implied warranty of MERCHANTABILITY or FITNESS FOR A     *
 *                  PARTICULAR PURPOSE. See the GNU LGPL for more details.   *
 *                                                                           *
 * You should have received a copy of the GNU Lesser General Public License  *
 * along with this library; if not, write to the Free Software Foundation,   *
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA          *
 * The GNU LEsser General Public License is contained in the file COPYING.   *
 *                                 ---------                                 *
 *   Barcelona Supercomputing Center - Centro Nacional de Supercomputacion   *
\*****************************************************************************/

/**
 *  @file    statistics.c
 *
 *  @brief   Structures to save statistics and show final briefing.
 *
 *  @author  Pedro A. Gonzalez - CEPBA-tools
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "statistics.h"
#include "log.h"

Record_Name_t otf_names[] =
{
	{"VERSION"},
	{"CREATOR"},
  {"COMMENT"},
	{"TIMER RESOLUTION"},
	{"PROCESS GROUP"},
	{"PROCESS"},
	{"FUNCTION GROUP"},
	{"FUNCTION"},
	{"COUNTER GROUP"},
	{"COUNTER"},
	{"COLLECTIVE OPERATION"},
  {"SCL"},
  {"SCL FILE"},
  {"EVENT COMMENT"},
	{"ENTER"},
	{"LEAVE"},
	{"SEND"},
	{"RECEIVE"},
	{"COUNTER"},
	{"COLLECTIVE OPERATION"},
  {"BEGIN PROCESS"},
  {"END PROCESS"},
  {"SUMMARY CONTENT"},
  {"FUNCTION SUMMARY"},
  {"FUNCTION GROUP SUMMARY"},
  {"MESSAGE SUMMARY"},
  {"SNAPSHOT COMMENT"},
  {"ENTER SNAPSHOT"},
  {"SEND SNAPSHOT"}
};


Record_Name_t prv_names[] =
{
	{"EVENTS"},
	{"COMMUNICATIONS"}
};


Record_Name_t queue_names[] =
{
	{"UNMATCHED SENDS"},
	{"UNMATCHED RECEIVES"},
	{"EVENTS"}
};


void Init_Statistics( Statistics_t *st )
{
  int ii;

  st->otf_num_records_read = 0;
  for ( ii = 0; ii < OTF_RECORD_TYPES; ii++ )
    st->otf_records[ ii ] = 0;

  st->otf_unmatched_sends = 0;
  st->otf_unmatched_receives = 0;
  st->otf_matched_sends = 0;
  st->otf_matched_receives = 0;

  st->prv_num_records_wrote = 0;
  for ( ii = 0; ii < PRV_RECORD_TYPES; ii++ )
    st->prv_records[ ii ] = 0;

  st->prv_negative_comms = 0;


  for ( ii = 0; ii < QUEUE_TYPES; ii++ )
  {
    st->queue[ii].queued   = 0;
    st->queue[ii].dequeued = 0;
    st->queue[ii].total_sum_queue = 0;
  }
}


static void Fill_Convert_Int( char *result, unsigned int num )
{
  char aux[16];
  char *offset;

  strcpy( result, "..............." );
  sprintf( aux, " %i", num);
  offset = result + ( 16 - strlen(aux) - 1);
  strcpy( offset, aux );
}


static void Fill_Convert_Msg( char *result, const char *msg, unsigned int spaces )
{
  int ii;
  char *offset;

  for ( ii = 0; ii < spaces; ii ++ )
    result[ ii ] = ' ';

  offset = result + (sizeof(char) * spaces);
  strcpy( offset, msg );
  ii = strlen( msg ) + spaces;
  while ( ii < 35 )
  {
    result[ ii ] = '.';
    ii++;
  }
  result[ ii ] = '\0';
}


static void Print_Msg_Int( unsigned int spaces,
                            const char *msg,
                            unsigned int num,
                            char *msg_buffer,
                            char *num_buffer )
{
  char line[ MAX_LINE ];
  
  Fill_Convert_Msg( msg_buffer, msg, spaces );
  Fill_Convert_Int( num_buffer, num );
  sprintf( line, "%s%s\n", msg_buffer, num_buffer );
  Write_Log( log_fd, ALWAYS, line );
}


void Print_Statistics( Statistics_t *st,
											 TranslatorOptions_t *param,
                       int argv,
											 char *argc[] )
{
  int ii;
  char *mbuf; /* message buffer */
  char *nbuf; /* number  buffer */
  float avg_queue_length;
  char line[ MAX_LINE ];
	
  nbuf = ( char * )malloc( sizeof( char ) * 16 );
  if ( nbuf == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  mbuf = ( char * )malloc( sizeof( char ) * 36 );
  if ( mbuf == NULL )
  {
    printf( "Out of Memory in %s:%i.\n", __FUNCTION__, __LINE__ );
    exit( 1 );
  }

  Write_Log( log_fd, ALWAYS, "\n\notf2prv statistics:\n" );
  
  for ( ii = 0; ii < argv; ii++ )
  {
    sprintf( line, "  %s", argc[ii] );
    Write_Log( log_fd, ALWAYS, line );
  }

  Write_Log( log_fd, ALWAYS, "\n" );

  sprintf( line, "  otf trace name... %s\n", param->otf_trace_name );
  Write_Log( log_fd, ALWAYS, line );
  sprintf( line, "  prv trace name... %s\n\n", param->prv_trace_name );
  Write_Log( log_fd, ALWAYS, line );

  /* DEFINITION */
  for ( ii = OTF_INI_DEF_REC; ii <= OTF_END_DEF_REC; ii ++ )
    st->otf_num_records_read += st->otf_records[ ii ];

  Print_Msg_Int( 2, "OTF total definition records read",
                    st->otf_num_records_read, mbuf, nbuf );

  for ( ii = OTF_INI_DEF_REC; ii <= OTF_END_DEF_REC; ii++ )
    Print_Msg_Int( 4, otf_names[ii].name, st->otf_records[ii], mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n" );

  /* EVENTS */
  st->otf_num_records_read = 0;

  for ( ii = OTF_INI_EVT_REC; ii <= OTF_END_EVT_REC; ii ++ )
    st->otf_num_records_read += st->otf_records[ ii ];

  Print_Msg_Int( 2, "OTF total events records read",
                    st->otf_num_records_read, mbuf, nbuf );
  for ( ii = OTF_INI_EVT_REC; ii <= OTF_END_EVT_REC; ii++ )
      Print_Msg_Int( 4, otf_names[ii].name, st->otf_records[ii], mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n" );

  /* SUMMARY */
  st->otf_num_records_read = 0;

  for ( ii = OTF_INI_SUM_REC; ii <= OTF_END_SUM_REC; ii ++ )
    st->otf_num_records_read += st->otf_records[ ii ];

  Print_Msg_Int( 2, "OTF total summary records read",
                    st->otf_num_records_read, mbuf, nbuf );
  for ( ii = OTF_INI_SUM_REC; ii <= OTF_END_SUM_REC; ii++ )
      Print_Msg_Int( 4, otf_names[ii].name, st->otf_records[ii], mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n" );

  /* SNAPSHOT */
  st->otf_num_records_read = 0;
  for ( ii = OTF_INI_SUM_REC; ii <= OTF_END_SUM_REC; ii ++ )
    st->otf_num_records_read += st->otf_records[ ii ];

  Print_Msg_Int( 2, "OTF total snapshot records read",
                    st->otf_num_records_read, mbuf, nbuf );
  for ( ii = OTF_INI_SNA_REC; ii <= OTF_END_SNA_REC; ii++ )
    Print_Msg_Int( 4, otf_names[ii].name, st->otf_records[ii], mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n" );

  /* MORE */
  Print_Msg_Int( 2, "OTF unmatched sends", st->otf_unmatched_sends, mbuf, nbuf );
  Print_Msg_Int( 2, "OTF unmatched receives", st->otf_unmatched_receives, mbuf, nbuf );
  Print_Msg_Int( 2, "OTF matched sends", st->otf_matched_sends, mbuf, nbuf );
  Print_Msg_Int( 2, "OTF matched receives", st->otf_matched_receives, mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n" );
  for ( ii = 0; ii < PRV_RECORD_TYPES; ii ++ )
    st->prv_num_records_wrote += st->prv_records[ ii ];
  Print_Msg_Int( 2, "PRV total records wrote", st->prv_num_records_wrote, mbuf, nbuf );
  for ( ii = 0; ii < PRV_RECORD_TYPES; ii++ )
    Print_Msg_Int( 4, prv_names[ii].name, st->prv_records[ii], mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n" );
  Print_Msg_Int( 2, "PRV negative communications", st->prv_negative_comms , mbuf, nbuf );

  Write_Log( log_fd, ALWAYS, "\n  QUEUES: ( queued, dequeued, average length )\n" );
  for ( ii = 0; ii < QUEUE_TYPES; ii++ )
  {
    if ( st->queue[ ii ].queued > 0 )
      avg_queue_length = ( float )(st->queue[ ii ].total_sum_queue /
                                   st->queue[ ii ].queued );
    else
      avg_queue_length = 0.0;
    sprintf( line, "    %s : ( %i, %i, %.2f )\n",
                     queue_names[ii].name,
                     st->queue[ii].queued,
                     st->queue[ii].dequeued,
                     avg_queue_length );
    Write_Log( log_fd, ALWAYS, line );
  }

  Write_Log( log_fd, ALWAYS, "\n\n");

  free( nbuf );
  free( mbuf );
}
